"""Various utility functions."""
from abc import abstractmethod
from dataclasses import dataclass
from typing import Protocol

import cv2 as cv
import numpy
import numpy as np
import qamlib
from typing_extensions import Self


class Comparable(Protocol):
    """Objects, whose values are comparable.

    This protocol is sometimes called Ord because it defines a total ordering
    of values of its type.

    Sometimes, the Ord protocol/interface is preceeded by a PartialOrd protocol
    for types that cannot be completely ordered. Here, we just ignore this
    distinction since Python and Numpy does the same.
    """

    @abstractmethod
    def __le__(self, other: Self) -> bool:
        """Is `self` less than or equal to `other`?"""  # noqa: D400
        pass

    @abstractmethod
    def __ge__(self, other: Self) -> bool:
        """Is `self` greater than or equal to `other`?"""  # noqa: D400
        pass

    @abstractmethod
    def __lt__(self, other: Self) -> bool:
        """Is `self` strictly less than `other`?"""  # noqa: D400
        pass

    @abstractmethod
    def __gt__(self, other: Self) -> bool:
        """Is `self` strictly greater than `other`?"""  # noqa: D400
        pass


def _validate_bits(int_bits: int, frac_bits: int, signed: bool):
    if int_bits < 0:
        ValueError("int_bits must be positive")

    if frac_bits < 0:
        ValueError("float_bits must be positive")

    if int_bits + frac_bits + int(signed) > 32:
        ValueError("(int_bits + frac_bits + 1) must be <= 32.")


def to_fixedpoint(value: float, int_bits: int, frac_bits: int, signed: bool = False):
    """Convert float value to fixed-point binary representation.

    The function is vectorized over `value`. The total number of bits for the fixed
    point result must be <= 32.

    :param value: The input values, either as a single float or as an array
        of floats.
    :param int_bits: The number of integer bits in the output representation.
    :param frac_bits: The number of fractional bits in the output representation.
    :param signed: Whether the output should be signed or not.
    :returns: The fixed point result, either as a single `int` or as an array of
        integers.
    """
    _validate_bits(int_bits, frac_bits, signed)

    # The total number of bits used for encoding the number (minus the sign bit)
    total_bits = int_bits + frac_bits - int(signed)
    # The maximum unsigned value that can be encoded
    max_val = 2**total_bits
    # Calculate the base number (which may be signed)
    base = int(value * 2**frac_bits)

    # Unsigned numbers are clamped to [0, max_val]
    if not signed:
        return max(0, min(max_val, base))

    if base >= 0:
        # Signed positive numbers are clamped to [0, max_val]
        return min(base, max_val - 1)
    else:
        # Signed negative numbers are clamped to [-max_val, 0)
        base = max(base, -max_val)
        # The sign bit is then added (max_val * 2)
        return base + max_val * 2


def from_fixedpoint(value: int, int_bits: int, frac_bits: int, signed: bool = False):
    """Convert fixed point binary values to floating point.

    The function is vectorized over `value`. The total number of bits for the fixed
    point result must be <= 32.

    :param value: The input values, either as a single integer or as an array
        of ints.
    :param int_bits: The number of integer bits in the output representation.
    :param frac_bits: The number of fractional bits in the output representation.
    :param signed: Whether the output should be signed or not.
    :returns: The floating point result, either as a single `float` or as an
        array of floats.
    """
    _validate_bits(int_bits, frac_bits, signed)

    # The total number of bits used for encoding the number (minus the sign bit)
    total_bits = int_bits + frac_bits - int(signed)
    top = 2**total_bits

    # The value is masked such that it doesn't include the sign bit.
    base = float(value & (top - 1))

    # Get the sign bit
    sign = signed and (value & top)

    if sign:
        # Negative numbers are simply negated before conversion
        return (base - top) / 2**frac_bits
    else:
        return base / 2**frac_bits


to_binary_vec = numpy.vectorize(
    to_fixedpoint, excluded=["int_bits", "frac_bits", "signed"]
)

from_binary_vec = numpy.vectorize(
    from_fixedpoint, excluded=["int_bits", "frac_bits", "signed"]
)


@dataclass
class Rectangle:
    """Rectangle utility class.

    :param p1: Lower left point.
    :param p2: Upper right point.
    """

    p1: np.ndarray
    p2: np.ndarray

    @property
    def center(self):
        """The rectangle center point."""
        return (self.p1 + self.p2) / 2

    @property
    def size(self):
        """The rectangle size."""
        return self.p2 - self.p1

    @property
    def p1_int(self):
        """Convert p1 to integer. Useful for drawing."""
        return np.int32(self.p1)

    @property
    def p2_int(self):
        """Convert p1 to integer. Useful for drawing."""
        return np.int32(self.p2)

    @classmethod
    def from_contour(cls, contour: np.ndarray) -> "Rectangle":
        """Create rectangle from contour points."""
        x, y, w, h = cv.boundingRect(contour)
        p1 = np.array([x, y])
        p2 = np.array([x + w, y + h])
        return cls(p1, p2)

    def scale_around_center(self, factor: float) -> "Rectangle":
        """Scale the rectangle around the center point.

        :param factor: Scaling factor. 1 means the rectangle is unchanged.
        :returns: A new scaled rectangle.
        """
        new_size = self.size * factor
        p1 = self.center - new_size / 2
        p2 = self.center + new_size / 2
        return Rectangle(p1, p2)

    def normalize(self, bounding_box: "Rectangle"):
        """Normalize rectangle with respect to a bounding box.

        The resulting rectangle's points will be relative to the bounding box, i.e.,
        point [0, 0] corresponds to the lower left point of `bounding_box` and [1, 1]
        corresponds to the upper right point.

        This method is useful when changing coordinate spaces.

        :param bounding_box: The reference rectangle.
        :returns: The scaled rectangle.
        """
        p1 = bounding_box._normalize_single(self.p1)
        p2 = bounding_box._normalize_single(self.p2)

        return Rectangle(p1, p2)

    def unnormalize(self, bounding_box: "Rectangle"):
        """Reverses the :meth:`Rectangle.normalize` function.

        Use this method to recover the original rectangle points for a normalized
        rectangle.

        :param bounding_box: The reference rectangle (with original points).
        :returns: The scaled rectangle.
        """
        p1 = bounding_box._unnormalize_single(self.p1)
        p2 = bounding_box._unnormalize_single(self.p2)

        return Rectangle(p1, p2)

    @property
    def points(self):
        """Get the four corner points of the rectangle."""
        return np.array(
            [
                [self.p1[0], self.p1[1]],
                [self.p1[0], self.p2[1]],
                [self.p2[0], self.p2[1]],
                [self.p2[0], self.p1[1]],
            ]
        )

    def _normalize_single(self, val: np.ndarray):
        return (val - self.p1) / self.size

    def _unnormalize_single(self, val: np.ndarray):
        return (val * self.size) + self.p1


def to_snake_case(name: str) -> str:
    """Convert a name from CamelCase to snake_case."""
    res = name[0].lower()
    for c in name[1:]:
        if c.isupper():
            res += "_" + c.lower()
        else:
            res += c
    return res


def cam_info_to_dict(info: dict[str, qamlib.Control]) -> dict[str, dict]:
    """Convert a `dict[str, qamlib.Control]` dictionary to a pure dictionary."""
    return {k: v.to_json() for k, v in info.items()}
